import "Turbine";
import "Turbine.Gameplay";
import "MinstrelBuffs.MinstrelBuffWindow.BuffWindow";
import "MinstrelBuffs.MinstrelBuffWindow.MinstrelEffects";
import "MinstrelBuffs.MinstrelBuffWindow.MinstrelBuffCommand";
import "MinstrelBuffs.MinstrelBuffWindow.Settings";
import "MinstrelBuffs.MinstrelBuffWindow.EventHelper";
import "MinstrelBuffs.MinstrelBuffWindow.OptionWindow";

LocalPlayer = Turbine.Gameplay.LocalPlayer.GetInstance();
EffectList = LocalPlayer:GetEffects();

MinstrelBuffMain = class();

-- Creates the main class.
function MinstrelBuffMain:Constructor()
	self.settings = MinstrelBuffs.MinstrelBuffWindow.Settings();
	self:RefreshMounted() -- Maybe Minstrelbuff is loaded when mounted.
	self.BuffWindow = MinstrelBuffs.MinstrelBuffWindow.BuffWindow(self.settings);
	self.optionWindow = OptionWindow(self.settings);
	self.balladsActive = false;

	-- Options support for Turbine PluginManager
	plugin.GetOptionsPanel = function() return self:GetOptionControl(); end
end

-- Writes all current buffs to the chat window.
function MinstrelBuffMain:WriteBuffsToChat()
	local placeCount = self:GetBuffCount();
	local effectCount = EffectList:GetCount();
	Turbine.Shell.WriteLine( "Current buffs: " .. placeCount );
	for	i = 1, effectCount, 1 do
		local effect = EffectList:Get(i);
		Turbine.Shell.WriteLine( effect:GetName() );	
	--[[
		if (i == 1) then
			effect.test = {};
			for k,v in pairs(effect) do
				Turbine.Shell.WriteLine( k .. " - " .. type(v));
				if (k == "__index" or k == "__metatable") then
					for k2,v2 in pairs(v) do
						Turbine.Shell.WriteLine( "   " .. k2 .. " - " .. type(v2));
						if (type(v2) == "table") then
							for k3,v3 in pairs(v2) do
								Turbine.Shell.WriteLine( "      " .. k3 .. " - " .. type(v3));
							end
						end
					end
				end
			end
		end
	end

	Turbine.Shell.WriteLine( "members" );
	for k,v in pairs(self) do
		Turbine.Shell.WriteLine( k .. "" );
	--]]
	end
end

-- Prepares all event handling and shows the buff window.
function MinstrelBuffMain:ShowWindow()
	self.EffectChangedCallback = function(sender, args) self:RefreshEffectDisplay(); end
	self.InCombatChanged = function(sender, args) self:RefreshVisibility(); end
	self.MountChanged = function(sender, args) self:RefreshMounted(); end
	self.MinstrelStanceChangedCallback = function(sender, args) self:MinstrelStanceChanged(); end
	self.CodaSkillResetTimeChangedCallback = function(sender) self:CodaSkillResetTimeChanged(); end
	self.CrySkillResetTimeChangedCallback = function(sender) self:CrySkillResetTimeChanged(); end
	self.MajorBalladResetTimeChangedCallback = function(sender) self:MajorBalladResetTimeChanged(); end
	self.codaWasPlayed = false;
	self.codaSkillList = { };
	self.crySkillList = { };
	self.majorBalladSkillList = { };
	
	-- Add buff, combat and warsteed events.
	AddCallback(EffectList, "EffectAdded", self.EffectChangedCallback);
	AddCallback(EffectList, "EffectRemoved", self.EffectChangedCallback);
	AddCallback(EffectList, "EffectCleared", self.EffectChangedCallback);
	AddCallback(LocalPlayer, "InCombatChanged", self.InCombatChanged);
	AddCallback(LocalPlayer, "MountChanged", self.MountChanged);
	if (LocalPlayer:GetClass() == Turbine.Gameplay.Class.Minstrel) then
		-- Add stance events
		self.classAttributes = LocalPlayer:GetClassAttributes();
		AddCallback(self.classAttributes, "StanceChanged", self.MinstrelStanceChangedCallback);
		
		-- Add coda events.
		local skillList = LocalPlayer:GetTrainedSkills();
		local skillCount = skillList:GetCount();
		--Turbine.Shell.WriteLine( "Current trained skills: " .. skillCount );
		local codaSkillListIndex = 1;
		local crySkillListIndex = 1;
		local majorBalladSkillListIndex = 1;
		for	i = 1, skillCount, 1 do
			local skill = skillList:GetItem(i);
			if (self:IsValidCodaSkill(skill)) then
				self.codaSkillList[codaSkillListIndex] = skill;
				codaSkillListIndex = codaSkillListIndex + 1;
				AddCallback(skill, "ResetTimeChanged", self.CodaSkillResetTimeChangedCallback);
			end

			if (self:IsValidCrySkill(skill)) then
				self.crySkillList[crySkillListIndex] = skill;
				crySkillListIndex = crySkillListIndex + 1;
				AddCallback(skill, "ResetTimeChanged", self.CrySkillResetTimeChangedCallback);
			end

			if (self:IsValidMajorBalladSkill(skill)) then
				self.majorBalladSkillList[majorBalladSkillListIndex] = skill;
				majorBalladSkillListIndex = majorBalladSkillListIndex + 1;
				AddCallback(skill, "ResetTimeChanged", self.MajorBalladResetTimeChangedCallback);
			end
		end		
	end

	-- Add event for position changing.
	self.PositionChangedCallback = function(sender, args) self:SaveMainPositionSettings(); end
	AddCallback(self.BuffWindow, "PositionChanged", self.PositionChangedCallback);

	-- Show the window.
	self.BuffWindow:Start();
	self:RefreshEffectDisplay();
end

-- Removes all events and hides the buff window.
function MinstrelBuffMain:HideWindow()
	if (self.EffectChangedCallback ~= nil) then
		RemoveCallback(EffectList, "EffectAdded", self.EffectChangedCallback);
		RemoveCallback(EffectList, "EffectRemoved", self.EffectChangedCallback);
		RemoveCallback(EffectList, "EffectCleared", self.EffectChangedCallback);
		RemoveCallback(LocalPlayer, "InCombatChanged", self.InCombatChanged);
		RemoveCallback(LocalPlayer, "MountChanged", self.MountChanged);
		if (self.classAttributes ~= nil) then
			RemoveCallback(self.classAttributes, "StanceChanged", self.MinstrelStanceChangedCallback);
			self.classAttributes = nil;
		end
		for i = 1, table.getn(self.codaSkillList), 1 do
			RemoveCallback(self.codaSkillList[i], "ResetTimeChanged", self.CodaSkillResetTimeChangedCallback);
		end
		for i = 1, table.getn(self.crySkillList), 1 do
			RemoveCallback(self.crySkillList[i], "ResetTimeChanged", self.CrySkillResetTimeChangedCallback);
		end
		for i = 1, table.getn(self.majorBalladSkillList), 1 do
			RemoveCallback(self.majorBalladSkillList[i], "ResetTimeChanged", self.MajorBalladResetTimeChanged);
		end
		self.EffectChangedCallback = nil;
	end	
	
	if (self.PositionChangedCallback ~= nil) then
		RemoveCallback(self.BuffWindow, "PositionChanged", self.PositionChangedCallback);
		self.PositionChangedCallback = nil;
	end
	self.BuffWindow:Stop();
end

-- Locks or unlocks the ui for positioning (CTRL + #)
-- locked - boolean
function MinstrelBuffMain:UiLocked(locked)
	self:SaveMainPositionSettings();
	self.BuffWindow:UiLocked(locked);
end

-- Opens the options window (in the turbine plugin manager)
function MinstrelBuffMain:ShowOptions()
	Turbine.PluginManager.ShowOptions(Plugins.MinstrelBuff);
end

-- Returns the options control. Used by turbines plugin manager.
function MinstrelBuffMain:GetOptionControl()
	return self.optionWindow:GetOptionControl();
end

-- Saves the position of the window.
function MinstrelBuffMain:SaveMainPositionSettings()
	--Turbine.Shell.WriteLine("SavePosition");
	if (self.BuffWindow ~= nil) then
		local posx, posy = self.BuffWindow:GetPosition();
		self.settings:SetMainPos(posx, posy);		
	end
end

-- Returns the localized effect names.
function MinstrelBuffMain:GetLocalization()
	return self.settings:GetEffectLocalizationNames();
end

-- Occurs, when a buff effect is added or removed.
-- Refreshes the BuffWindow contents.
function MinstrelBuffMain:RefreshEffectDisplay()
	local effectCount = EffectList:GetCount();
	local balladEffectPlace = 1;
	local anthemEffectPlace = 1;
	local sortedEffects = {};
	local isAnthemFixed = self.settings:GetFixedAnthemPositions();
	local showAnthems = self.settings:GetShowAnthems();
	for	i = 1, effectCount, 1 do
		sortedEffects[i] = EffectList:Get(i);
	end
	table.sort(sortedEffects, function(a,b) return a:GetStartTime() < b:GetStartTime() end);
	
	if (isAnthemFixed == true) then
		for i = 1, 6, 1 do
			self.BuffWindow:HideAnthem(i);
		end
	end

	for	i = 1, effectCount, 1 do
		--local effect = EffectList:Get(i);
		local effect = sortedEffects[i];
		if (balladEffectPlace <= 3 and self:IsValidBalladEffect(effect:GetName())) then
			self.BuffWindow:ShowEffect(effect, balladEffectPlace);
			balladEffectPlace = balladEffectPlace + 1;
		end
		if (showAnthems == true) then
			if (isAnthemFixed == false) then
				if (anthemEffectPlace <= 6 and self:IsValidAnthemEffect(effect:GetName())) then
					self.BuffWindow:ShowAnthem(effect, anthemEffectPlace);
					anthemEffectPlace = anthemEffectPlace + 1;
				end
			else
				if (self:IsValidAnthemEffect(effect:GetName())) then
					local anthemEffectPosition = self:GetAnthemEffectPosition(effect:GetName());
					self.BuffWindow:ShowAnthem(effect, anthemEffectPosition);
					-- Effect border only works with fixed anthems.
					self.BuffWindow:ShowAnthemBorder(anthemEffectPosition);
				end
			end
		end
	end

	if (balladEffectPlace == 1 and self.balladsActive) then
		-- Went from having ballads to not having ballads
		self.balladsActive = false;
		-- Make sure the combat timer stopped
		self.BuffWindow:BalladEnded();
	elseif (balladEffectPlace > 1 and not self.balladsActive) then
		-- Went from not having ballads to having ballads
		self.balladsActive = true;
	end
	
	if (balladEffectPlace <= 3) then
		for i = balladEffectPlace, 3, 1 do
			self.BuffWindow:HideEffect(i);
		end
		if (balladEffectPlace == 1 and self.codaWasPlayed == true) then -- Coda was played, hide all borders.
			self.codaWasPlayed = false;
			self.BuffWindow:HideAllAnthemBorders();
		end
	end
	if (anthemEffectPlace <= 6 and isAnthemFixed == false) then
		for i = anthemEffectPlace, 6, 1 do
			self.BuffWindow:HideAnthem(i);
		end
	end
	--[[
	local party = LocalPlayer:GetParty();
	if (party ~= nil and party:GetMemberCount() > 1) then
		Turbine.Shell.WriteLine("Count: " .. party:GetMemberCount());
		local partyPlayer = party:GetMember(1);
		Turbine.Shell.WriteLine("Name: " .. partyPlayer:GetName());
		Turbine.Shell.WriteLine("Morale: " .. partyPlayer:GetMorale());
		--Turbine.Shell.WriteLine("Effects: " .. partyPlayer:GetEffects());
		
		--local partyEffectList = partyPlayer.EffectList;
		--DebugTable(party);		
		local partyEffectlist = partyPlayer:GetEffects();
		if (partyEffectlist ~= nil) then
			Turbine.Shell.WriteLine("Effects: " .. partyEffectlist:GetCount());
			--DebugTable(EffectList);
		end
		--local partyEffectCount = partyEffectList:GetCount();
		--Turbine.Shell.WriteLine("PartyEffects: " .. partyEffectCount);
	end
	]]--
end

function DebugTable(class)
	if type(class) ~= "table" then
	 Turbine.Shell.WriteLine( tostring(class) .. ": (" .. type(class) .. ")");
	 return;
	end
  local sortAtable = {}
  for key,v in pairs(class) do table.insert(sortAtable, {key,v}) end
  table.sort(sortAtable,compareItems);
	Turbine.Shell.WriteLine( tostring(class) .. ": (" .. type(class) .. ") " .. " : # " .. #sortAtable);
	cltb = getmetatable(class);
	for n,var in pairs(sortAtable) do
	 k = var[1]; v = var[2];
	 Turbine.Shell.WriteLine( "(" .. k .. "): (" .. type(v) .. ") " .. tostring(v) .. " ;" );
	 if k == "__implementation" then
	  u = getmetatable(v);	v3 = v;
	  if u then
	   for k2, v2 in pairs( u ) do
	    Turbine.Shell.WriteLine( "C: (" .. k2 .. "): " .. tostring(v2) );
	    if k2 == "__index" then v3 = v2; end
	   end
	   if v3 == v then
	    if u.__index then v3 = u.__index;
	    elseif cltb and type(v) == "userdata" then
	     if cltb.__index then v3 = cltb.__index;
	     else v3 = cltb; end
	    end
	   end
	  end
	  u = v3;
	 elseif k == "__index" and not cltb and getmetatable(v) then u = getmetatable(v).__index;
	 else u = v; end
--	 Turbine.Shell.WriteLine( "(" .. k .. "): (" .. type(u) .. ") " .. tostring(u) .. " ;" );
	 if type(u) == "table" then  -- can call DebugTable(u) recursively instead, but it's not advisable
	  table.sort(u);
	  for k2, v2 in pairs( u ) do
	    Turbine.Shell.WriteLine( "(" .. k2 .. "): " .. tostring(v2) );
	  end
	 elseif u ~= v then
	  Turbine.Shell.WriteLine( tostring(u) );
	 end
	end
	if cltb then
	 for k2, v2 in pairs( cltb ) do
	  Turbine.Shell.WriteLine( "  (" .. k2 .. "): " .. tostring(v2) );
	 end
	end
end

-- Returns true, if the effect is a valid ballad effect.
-- effectName - string
function MinstrelBuffMain:IsValidBalladEffect(effectName)
	if (self:GetLocalization() == "fr") then
		return ValidBalladEffectsFr[effectName] ~= nil;
	elseif (self:GetLocalization() == "de") then
		return ValidBalladEffectsDe[effectName] ~= nil;
	else
		return ValidBalladEffectsEn[effectName] ~= nil;
	end
end

-- Returns true, if the effect is a valid anthem effect.
-- effectName - string
function MinstrelBuffMain:IsValidAnthemEffect(effectName)
	if (self:GetLocalization() == "fr") then
		return ValidAnthemEffectsFr[effectName] ~= nil;
	elseif (self:GetLocalization() == "de") then
		return ValidAnthemEffectsDe[effectName] ~= nil;
	else
		return ValidAnthemEffectsEn[effectName] ~= nil;
	end
end

-- Returns the position of the effect.
-- effectName - string
function MinstrelBuffMain:GetAnthemEffectPosition(effectName)
	if (self:GetLocalization() == "fr") then
		return ValidAnthemEffectsFr[effectName];
	elseif (self:GetLocalization() == "de") then
		return ValidAnthemEffectsDe[effectName];
	else
		return ValidAnthemEffectsEn[effectName];
	end
end

-- Returns true, if the skill is a minstrel coda.
-- skill - Turbine.Gameplay.ActiveSkill
function MinstrelBuffMain:IsValidCodaSkill(skill)
	local skillInfo = skill:GetSkillInfo();
	local skillName = skillInfo:GetName();
	if (self:GetLocalization() == "fr") then
		return ValidCodaSkillNamesFr[skillName] ~= nil;
	elseif (self:GetLocalization() == "de") then
		return ValidCodaSkillNamesDe[skillName] ~= nil;
	else
		return ValidCodaSkillNamesEn[skillName] ~= nil;
	end
end

function MinstrelBuffMain:IsValidCrySkill(skill)
	local skillInfo = skill:GetSkillInfo();
	local skillName = skillInfo:GetName();
	if (self:GetLocalization() == "fr") then
		return ValidCrySkillNamesFr[skillName] ~= nil;
	elseif (self:GetLocalization() == "de") then
		return ValidCrySkillNamesDe[skillName] ~= nil;
	else
		return ValidCrySkillNamesEn[skillName] ~= nil;
	end
end

function MinstrelBuffMain:IsValidMajorBalladSkill(skill)
	local skillInfo = skill:GetSkillInfo();
	local skillName = skillInfo:GetName();
	if (self:GetLocalization() == "fr") then
		return ValidMajorBalladSkillNamesFr[skillName] ~= nil;
	elseif (self:GetLocalization() == "de") then
		return ValidMajorBalladSkillNamesDe[skillName] ~= nil;
	else
		return ValidMajorBalladSkillNamesEn[skillName] ~= nil;
	end	
end

-- Gets the count of the buffs.
function MinstrelBuffMain:GetBuffCount()
	local count = 0;
	count = EffectList:GetCount();
	return count;
end

-- Returns true, if the player is currently on a war steed.
function MinstrelBuffMain:IsPlayerOnWarsteed()
    local mount = LocalPlayer:GetMount();
    return mount ~= nil and mount:IsA(Turbine.Gameplay.CombatMount);
end

-- Occurs, when the minstrel stance was changed.
function MinstrelBuffMain:MinstrelStanceChanged()
	self.BuffWindow:SetAnthemBackgroundFromStance();
	--Turbine.Shell.WriteLine(self.classAttributes:GetStance());
end

-- Occurs, when a coda skill was used.
function MinstrelBuffMain:CodaSkillResetTimeChanged()
	self.codaWasPlayed = true; -- only mark coda as played, because the resetting event for the anthems and ballads can occur later.
	--self.BuffWindow:HideAllAnthemBorders();
end

-- Occurs, when a cry skill was used.
function MinstrelBuffMain:CrySkillResetTimeChanged()
	-- Tell the window about the event:
	self.BuffWindow:CryOfTheChorus();
end

-- Occurs, when a major ballad skill was used.
function MinstrelBuffMain:MajorBalladResetTimeChanged()
	-- Tell the window about the event:
	self.BuffWindow:MajorBallad();
end

-- Refreshes the visibility of the main window (i.e. when out of combat, no effects displayed).
function MinstrelBuffMain:RefreshVisibility()
	self.BuffWindow:SetVisibility();
end

-- Refreshes the window when the player is on a war steed or gets off a war steed.
function MinstrelBuffMain:RefreshMounted()
	self.settings:SetIsOnWarsteed(self:IsPlayerOnWarsteed());
end

mbMain = MinstrelBuffMain();
--mbCommand = MinstrelBuffs.MinstrelBuffWindow.MinstrelBuffCommand(mbMain);
MinstrelBuffCommand.MinstrelBuffMain = mbMain;
Turbine.Shell.AddCommand( "minstrelbuff", MinstrelBuffCommand);
Turbine.Shell.AddCommand( "mb", MinstrelBuffCommand);

mbMain:ShowWindow();
MinstrelBuffCommand:ShowHelpText();